﻿using System;
using System.Configuration;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Drawing;
using System.Globalization;
using System.IO;
using System.Linq;
using System.ServiceModel;
using System.ServiceModel.Security;
using System.Windows.Forms;
using EmmaPortal.Business.Utility;
using log4net;
using SubscriberServicesClient.Emma.Subscription.Service.Client;
using SubscriberServicesClient.Helpers;
using log4net.Config;

namespace SubscriberServicesClient
{
	public partial class Form1 : Form, ISubscriberServiceForm
	{
		public static readonly object[] SupportedSubmissionTypes = { new SubmissionType("Continuing Disclosure", "CD"), new SubmissionType("Financial Reporting", "FR"), new SubmissionType("Primary Market", "PM"), new SubmissionType("Presale", "PS"), new SubmissionType("Short Docs", "AV"), new SubmissionType("Invalid", "C4") };

		public Action<Cursor> SetCursorType = SetSystemCursor;
		public Func<IEmmaSubscriptionService> GetEmmaDataService;
		public Func<string, DialogResult> ShowMessageBox;
		public Action<bool> EnableFileRetrievalButton;
		public Func<string, Process> StartProcess = Process.Start;
		public Func<string, Stream> GetTemporaryStream = GetTemporaryFileStream;
		public Func<string> GetTemporaryFileName = GetTemporaryFileNameFromPath;
		public Func<Point, int> GetTransactionIndexFromPoint;

		private readonly ISystemClock _clock;
		private readonly ILog _logger;
		private bool _shouldUseGatewayCredentials;
		private string _selectedCertificateName;
		public Action CloseCurrentConnection;

		[ExcludeFromCodeCoverage]
		public Form1() : this(new SystemClock(), LogManager.GetLogger("ERROR"))
		{
		}

		public Form1(ISystemClock clock, ILog logger)
		{
			XmlConfigurator.Configure();
			GetEmmaDataService = GetEmmaDataServiceClient;
			CloseCurrentConnection = CloseEmmaDataServiceConnection;
			ShowMessageBox = MessageBox.Show;
			EnableFileRetrievalButton = state => btnGetFile.Enabled = state;
			_logger = logger;
			_clock = clock;
			InitializeComponent();
			GetTransactionIndexFromPoint = point => lstTransactions.IndexFromPoint(point);
		}

		public ListBox TransactionList
		{
			get
			{
				return lstTransactions;
			}
		}

		public DateTime SelectedPostingDate
		{
			set
			{
				dtpPostingDate.Value = value;
			}
		}

		[ExcludeFromCodeCoverage]
		private static void SetSystemCursor(Cursor type)
		{
			Cursor.Current = type;
		}

		public ComboBox SubmissionTypes
		{
			get
			{
				return ddlSubmissionType;
			}
		}

		public Label StatusMessage
		{
			get
			{
				return lRecordsReturned;
			}
		}

		public void GetTransactions(object sender, EventArgs e)
		{
			TransactionList.Items.Clear();
			try
			{
				SetCursorType(Cursors.WaitCursor);
				var response = GetEmmaDataService().GetTransactions(((SubmissionType)ddlSubmissionType.SelectedItem).Code, dtpPostingDate.Value, txtSpecVersion.Text);
				RequestId.Text = response.RequestId;

				if (response.StatusCode == 200)
				{
					TransactionList.Items.AddRange(response.Transactions.Select(AsListItem).ToArray());
					StatusMessage.Text = string.Format("{0} - Requested at {1}", TransactionList.Items.Count.ToString(), DateTime.Now.ToShortTimeString());
				}
				else
				{
					ShowMessageBox(string.Format("An unexpected error occurred during processing and has been logged. If the error persists contact the MSRB. Code: {0}.", response.StatusCode));
					StatusMessage.Text = ErrorMessageText;
				}

			}
			catch (NotLoggedInException)
			{
				ShowLoginMessage();
			}
			catch (Exception ex)
			{
				ShowMessageBox(string.Format("An unexpected error occurred during processing and has been logged. If the error persists contact the MSRB. Code: {0}", ex.Message));
				_logger.Error("Unable to retrieve transaction list.", ex);
				StatusMessage.Text = ErrorMessageText;
			}
			finally
			{
				SetCursorType(Cursors.Default);
			}
		}

		private void ShowLoginMessage()
		{
			ShowMessageBox("Please log in before trying this operation.");
		}

		private string ErrorMessageText
		{
			get
			{
				return string.Format("Error - Requested at {0}", _clock.Now.ToShortTimeString());
			}
		}

		public ComboBox EndpointAddress
		{
			get
			{
				return cbEndpoint;
			}
		}

		public EndpointAddress GetEndpointAddress()
		{
			return new EndpointAddress(new Uri(EndpointAddress.Text));
		}

		private static string AsListItem(EmmaTransaction transaction)
		{
            return string.Format("{0}; posted {1}", transaction.SubmissionId, transaction.PostingDateTime.ToString("MM/dd/yyyy HH:mm:ss.fff", CultureInfo.InvariantCulture));
		}

		public void Form1_Load(object sender, EventArgs e)
		{
			SelectedPostingDate = _clock.Today;
			InitializeSubmissionTypes();
			ClearLoginInformation();
			InitializeEndpoints();
		}

		private void ClearLoginInformation()
		{
			CurrentUserName = string.Empty;
		}

		private void InitializeEndpoints()
		{
			var path = ConfigurationManager.AppSettings["ServicePath"];
			foreach (var key in ConfigurationManager.AppSettings.AllKeys)
			{
				if (key.StartsWith("Server"))
				{
					cbEndpoint.Items.Add(string.Format("https://{0}/{1}", ConfigurationManager.AppSettings[key], path));
				}
			}
			cbEndpoint.SelectedIndex = 0;
		}

		private void InitializeSubmissionTypes()
		{
			ddlSubmissionType.Items.Clear();
			ddlSubmissionType.Items.AddRange(SupportedSubmissionTypes);
			ddlSubmissionType.SelectedIndex = 0;
		}

		public class SubmissionType
		{
			public SubmissionType(string name, string code)
			{
				Name = name;
				Code = code;
			}

			public string Code { get; private set; }

			private string Name { get; set; }

			public override string ToString()
			{
				return Name;
			}
		}

		public TextBox FileId
		{
			get
			{
				return txtFileId;
			}
		}

		[ExcludeFromCodeCoverage]
		private static string GetTemporaryFileNameFromPath()
		{
			return Path.ChangeExtension(Path.GetTempFileName(), "pdf");
		}

		[ExcludeFromCodeCoverage]
		private static Stream GetTemporaryFileStream(string fileName)
		{
			return new FileStream(fileName, FileMode.Create);
		}

		public void GetFile(object sender, EventArgs e)
		{
			try
			{
				EnableFileRetrievalButton(false);
				SetCursorType(Cursors.WaitCursor);
				var file = GetEmmaDataService().GetFile(txtFileId.Text);
				var fileName = GetTemporaryFileName();
				CreateTemporaryFile(fileName, file);
				StartProcess(fileName);
			}
			catch (NotLoggedInException)
			{
				ShowLoginMessage();
			}
			catch (FaultException<FileRetrievalFault> ex)
			{
				ShowMessageBox(string.Format("FileRetrievalFault: FileId={0}, StatusCode={1}", ex.Detail.RequestedFileId, ex.Detail.StatusCode));
			}
			finally
			{
				EnableFileRetrievalButton(true);
				SetCursorType(Cursors.Default);
			}
		}

		private void CreateTemporaryFile(string fileName, Stream file)
		{
			using (file)
			{
				using (var output = GetTemporaryStream(fileName))
				{
					var buffer = new byte[8196];
					int length;
					while ((length = file.Read(buffer, 0, buffer.Length)) > 0)
					{
						output.Write(buffer, 0, length);
					}
				}
			}
		}

		public string CurrentEndpointName
		{
			[ExcludeFromCodeCoverage]
			get
			{
				return _currentEndpointName;
			}
			set
			{
				CloseCurrentConnection();
				_currentEndpointName = value;
			}
		}

		[ExcludeFromCodeCoverage]
		private void CloseEmmaDataServiceConnection()
		{
			if (_client != null)
			{
				_client.Close();
			}
			_client = null;
		}

		private EmmaSubscriptionServiceClient _client;
		private string _currentEndpointName;
		private string _currentPassword;

		[ExcludeFromCodeCoverage]
		private IEmmaSubscriptionService GetEmmaDataServiceClient()
		{
			if (_client != null)
			{
				return _client;
			}
			if (string.IsNullOrEmpty(CurrentEndpointName))
			{
				throw new NotLoggedInException();
			}
			_client = new EmmaSubscriptionServiceClient(CurrentEndpointName, GetEndpointAddress());
			if (ShouldUseGatewayCredentials)
			{
				_client.ClientCredentials.ServiceCertificate.Authentication.CertificateValidationMode = X509CertificateValidationMode.None;
				_client.ClientCredentials.UserName.UserName = CurrentUserName;
				_client.ClientCredentials.UserName.Password = CurrentPassword;
			}
			else
			{
				_client.ClientCredentials.ClientCertificate.Certificate = SelectedCertificateName == null ? null : CertificateRepository.GetSelectedCertificate(SelectedCertificateName);
			}
			return _client;
		}

		public Label LoginType
		{
			get
			{
				return lblLoginType;
			}
		}

		public bool ShouldUseGatewayCredentials
		{
			[ExcludeFromCodeCoverage]
			get
			{
				return _shouldUseGatewayCredentials;
			}
			set
			{
				CloseCurrentConnection();
				_shouldUseGatewayCredentials = value;
				LoginType.Text = value ? "Gateway Credentials" : "Certificate";
				EndpointAddress.Text = GetEndpointAddressUrl(value, EndpointAddress.Text);
				for (var index = 0; index < EndpointAddress.Items.Count; index++)
				{
					EndpointAddress.Items[index] = GetEndpointAddressUrl(value, (string)EndpointAddress.Items[index]);
				}
			}
		}

		private static string GetEndpointAddressUrl(bool useGatewayCredentials, string text)
		{
			var index = text.IndexOf(".svc");
			return index < 0 ? text : text.Substring(0, index) + ".svc/" + (useGatewayCredentials ? "subscriber" : "regulator");
		}

		public string CurrentPassword
		{
			[ExcludeFromCodeCoverage]
			get
			{
				return _currentPassword;
			}
			set
			{
				CloseCurrentConnection();
				_currentPassword = value;
			}
		}

		public string SelectedCertificateName
		{
			get
			{
				return _selectedCertificateName;
			}
			set
			{
				_selectedCertificateName = value;
				CurrentUserName = value;
			}
		}

		public void lstTransactions_MouseDoubleClick(object sender, MouseEventArgs e)
		{
			var index = GetTransactionIndexFromPoint(e.Location);
			if (index == ListBox.NoMatches)
			{
				return;
			}
			var parsed = lstTransactions.Items[index].ToString().Split(';');
			AddTab(lstTransactions.Items[index].ToString(), GetSubmissionInfo(((SubmissionType)ddlSubmissionType.SelectedItem).Code, parsed[0], Convert.ToDateTime(parsed[1].Replace("posted", string.Empty).Trim())));
		}

		private SingleEmmaTransaction GetSubmissionInfo(string submissionType, string submissionId, DateTime postingDate)
		{
			SingleEmmaTransaction response;
			try
			{
				SetCursorType(Cursors.WaitCursor);
				response = GetEmmaDataService().GetTransaction(submissionType, postingDate, submissionId, txtSpecVersion.Text);
				if (response.StatusCode != 200)
				{
					ShowMessageBox("An unexpected error occurred during processing and has been logged. If the error persists contact the MSRB. Code=" + response.StatusCode);
				}
			}
			catch (NotLoggedInException)
			{
				ShowLoginMessage();
				response = new SingleEmmaTransaction();
			}
			catch (Exception ex)
			{
				response = new SingleEmmaTransaction { ReturnData = ex.Message };
			}
			finally
			{
				SetCursorType(Cursors.Default);
			}

			return response;
		}

		private void AddTab(string title, SingleEmmaTransaction transaction)
		{
			var myTabPage = new TabPage(title)
								{
									Anchor = AnchorStyles.Bottom | AnchorStyles.Left | AnchorStyles.Right | AnchorStyles.Top,
									BorderStyle = BorderStyle.FixedSingle,
									AutoSize = true,
									Dock = DockStyle.Fill
								};
			tabControl1.TabPages.Add(myTabPage);
			tabControl1.SelectTab(myTabPage);
			var control = new TransactionDataControl(transaction);
			myTabPage.Controls.Add(control);
			control.ResizeContents(myTabPage.Size);
			control.Close += CloseButtonOnClick;
		}

		public void CloseButtonOnClick(object sender, EventArgs eventArgs)
		{
			var currentTab = tabControl1.SelectedIndex;
			tabControl1.TabPages.RemoveAt(currentTab);
			tabControl1.SelectedIndex = Math.Min(currentTab, tabControl1.TabCount - 1);
		}

		public TabControl TabContainer
		{
			get
			{
				return tabControl1;
			}
		}

		private void ddlSubmissionType_SelectedIndexChanged(object sender, EventArgs e)
		{
			lstTransactions.Items.Clear();
		}

		public TextBox RequestId
		{
			get
			{
				return txtRequestId;
			}
		}

		public string CurrentUserName
		{
			get
			{
				return lblUserName.Text;
			}
			set
			{
				CloseCurrentConnection();
				lblUserName.Text = value;
			}
		}

		[ExcludeFromCodeCoverage]
		public void ShowLoginDialog(object sender, EventArgs e)
		{
			new LoginDialog(this).ShowDialog();
		}

		public class NotLoggedInException : Exception
		{
		}

		public void Form1_FormClosing(object sender, FormClosingEventArgs e)
		{
			CloseCurrentConnection();
		}
	}
}
