﻿using System;
using System.Diagnostics.CodeAnalysis;
using System.Drawing;
using System.IO;
using System.Windows.Forms;
using System.Xml.Serialization;
using Newtonsoft.Json;
using Newtonsoft.Json.Converters;
using SubmissionModel.ContinuingDisclosure;
using SubmissionModel.FinancialReporting;
using SubmissionModel.PrimaryMarket;
using SubmissionModel.ShortDocs;
using SubscriberServicesClient.Emma.Subscription.Service.Client;

namespace SubscriberServicesClient
{
	public partial class TransactionDataControl : UserControl
	{
		public event EventHandler Close;

		public TransactionDataControl(SingleEmmaTransaction transaction)
		{
			InitializeComponent();
			txtRequestId.Text = transaction.RequestId;
			txtResponseData.Text = transaction.ReturnData;
			SubmissionType = transaction.RequestedSubmissionType;
			btnClose.Click += OnClose;
		}

		[ExcludeFromCodeCoverage]
		public TransactionDataControl() : this(new SingleEmmaTransaction())
		{
		}

		public Button SaveAsButton
		{
			get
			{
				return btnSaveAs;
			}
		}

		public virtual ISaveFileDialogWrapper SaveDialog
		{
			get
			{
				return new WinFormsSaveFileDialogWrapper(dlgSave);
			}
		}

		public Button CloseButton
		{
			get
			{
				return btnClose;
			}
		}

		public TextBox RequestId
		{
			get
			{
				return txtRequestId;
			}
		}

		public TextBox ResponseData
		{
			get
			{
				return txtResponseData;
			}
		}
		public string SubmissionType { get; set; }

		private void OnClose(object sender, EventArgs e)
		{
			if (Close == null)
			{
				return;
			}
			Close.Invoke(sender, e);
		}

		public void ResizeContents(Size size)
		{
			txtResponseData.Width = size.Width - 8;
			txtResponseData.Height = size.Height - btnClose.Height - 8;
			btnClose.Left = txtResponseData.Width - btnClose.Width;
		}

		private void SaveTransaction(object sender, EventArgs e)
		{
			if (SaveDialog.ShowDialog() != DialogResult.OK)
			{
				return;
			}
			using (var writer = new StreamWriter(SaveDialog.OpenFile()))
			{
				writer.Write(GetTextForOutput(Path.GetExtension(SaveDialog.FileName)));
			}
		}

		private string GetTextForOutput(string type)
		{
			if (type == ".json")
			{
				return JsonConvert.SerializeObject(GetJsonWrapper(), Formatting.Indented, new StringEnumConverter());
			}
			return txtResponseData.Text;
		}

		private object GetJsonWrapper()
		{
			switch (SubmissionType)
			{
				case "CD":
					return new ContinuingDisclosureWrapper
					       {
							   ContinuingDisclosureSubmission = new XmlSerializer(typeof(ContinuingDisclosureSubmissionDetailsType)).Deserialize(new StringReader(txtResponseData.Text))
					       };
				case "FR":
					return new FinancialReportingWrapper
					{
						FinancialReportingSubmission = new XmlSerializer(typeof(FinancialReportingSubmissionDetailsType)).Deserialize(new StringReader(txtResponseData.Text))
					};
				case "AV":
					return new GenericSubmissionWrapper
					{
						Submission = new XmlSerializer(typeof(ShortSubmissionDetailsType)).Deserialize(new StringReader(txtResponseData.Text))
					};
				default:
					return new GenericSubmissionWrapper
					       {
						       Submission = new XmlSerializer(typeof(SubmissionDetailsType)).Deserialize(new StringReader(txtResponseData.Text))
					       };
			}
		}

		public class GenericSubmissionWrapper
		{
			public object Submission { get; set; }
		}

		public class ContinuingDisclosureWrapper
		{
			public object ContinuingDisclosureSubmission { get; set; }
		}

		public class FinancialReportingWrapper
		{
			public object FinancialReportingSubmission { get; set; }
		}
	}
}
